// SPDX-License-Identifier: MIT
pragma solidity ^0.8.20;

interface IClaim {
    error NotEnoughVested();
    error ClaimedOverMax();
    error PercentEmpty();
    error WalletNotPushed();
    error AddressExists();
    error TooMuch();
    error NotAnOwner();
    error NotADao();

    struct Term {
        uint256 percent;
        uint256 claimed;
        uint256 gClaimed;
        uint256 max;
    }

    function terms(address _address) external view returns (Term memory);
}
